#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu

# kill subprocesses when the caller script terminates
trap "trap - SIGTERM && jobs -p | xargs -r kill" INT TERM EXIT

#GLOBALS:
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error_with_code "10001"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error_with_code "10002"
fi

readonly SCRIPT_DIR
readonly MODULES_DIR

export OFFLINE_TMP_FOLDER="/tmp"
dev_null="/dev/null"

## Locale
LANG_FABRIC_STARTED_PREFIX="[ManagementTools] [Started]"
LANG_FABRIC_FAILED_PREFIX="[ManagementTools] [Failed]"
LANG_FABRIC_COMPLETED_PREFIX="[ManagementTools] [Completed]"
LANG_FABRIC_ALM_SETUP_MESSAGE="Installing and configuring Application Lifecycle Management Tools"
LANG_FABRIC_CLUSTER_TOOLS_SETUP_MESSAGE="Installing and configuring Cluster Management Tools"
LANG_SELF_HEAL_FAILED_PREFIX="[Self Heal Operator Installation] [Failed]"
SELF_HEAL_CHART="helm/self-heal-operator"
TMP_HELM_CHARTS_DIR="/tmp/sf-installer/fabric/selfheal/helm"

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/utils.sh"
# shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/longhornutils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/argocd_utils.sh"
#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}/cluster_utils.sh"

declare -a additional_parameters

# Ceph downgrade variable
TMP_CEPH_BACKUP_PATH=${TMP_CEPH_BACKUP_PATH:-""}

# Check if ceph cluster not using version: 16.2.6 due to bug https://docs.ceph.com/en/quincy/releases/pacific/#v16-2-6-pacific
function is_ceph_version_stable() {
  local return_code=0
  if kubectl -n rook-ceph get deployment -l rook_cluster  -o jsonpath='{range .items[*]}{.metadata.labels.ceph-version}{"\n"}{end}' | sort | uniq | grep -q '^16\.2\.6'; then
    return_code=1
  fi
  return "${return_code}"
}

function get_ceph_object_size_bytes() {
  local ceph_size
  ceph_size=$(kubectl -n rook-ceph exec deploy/rook-ceph-tools -- ceph status --format json | jq -r '.pgmap.data_bytes')
  echo "${ceph_size}" | grep -q -E '^[0-9]+$' || error "Invalid value: '${ceph_size}' ceph object size"
  echo "${ceph_size}"
}

function validate_ceph_tmp_backup_path() {
  local ceph_size="$1"
  local backup_path_available_capacity

  [[ -n "${TMP_CEPH_BACKUP_PATH}" ]] || error "Ceph cluster is running on version: 16.2.6 with known bug(https://docs.ceph.com/en/quincy/releases/pacific/#v16-2-6-pacific). Please provide temporary backup directory (via environment variable 'TMP_CEPH_BACKUP_PATH') to downgrade ceph"
  [[ -d "${TMP_CEPH_BACKUP_PATH}" ]] || error "Ceph cluster is running on version: 16.2.6 with known bug(https://docs.ceph.com/en/quincy/releases/pacific/#v16-2-6-pacific). Please provide valid temporary backup directory (via environment variable 'TMP_CEPH_BACKUP_PATH') to downgrade ceph"
  backup_path_available_capacity=$(findmnt --raw --bytes  -n -o AVAIL  --target "${TMP_CEPH_BACKUP_PATH}")
  [[ "$backup_path_available_capacity" -ge "$ceph_size" ]] || error "Ceph backup path: '${TMP_CEPH_BACKUP_PATH}' does not have enough space available to backup ceph objects of size: '$(numfmt --to=iec-i "$ceph_size")'"
  ! [[ -e "${TMP_CEPH_BACKUP_PATH}/backup" && ! -d "${TMP_CEPH_BACKUP_PATH}/backup" ]] || error "Path: ${TMP_CEPH_BACKUP_PATH}/backup must be of type: directory"
  if [[ -d "${TMP_CEPH_BACKUP_PATH}/backup" ]]; then
    [[ "$(stat -c '%U'  "${TMP_CEPH_BACKUP_PATH}/backup")" != "nobody" ]]  && error "Owner of folder ${TMP_CEPH_BACKUP_PATH}/backup must be nobody"
    [[ -z "$(find "${TMP_CEPH_BACKUP_PATH}/backup" ! -user nobody)" ]] || error "Directory: ${TMP_CEPH_BACKUP_PATH}/backup must not contain file/folder with user owner other than: 'nobody'"
  fi
}

function prepare_ceph_tmp_backup_path() {
  mkdir -p "${TMP_CEPH_BACKUP_PATH}"
  chmod 777 "${TMP_CEPH_BACKUP_PATH}"
}

function prepare_ceph_tmp_backup_script_configmap() {
  kubectl -n rook-ceph create configmap pre-upgrade-ceph-objectstore-backup-restore-cm --from-file "${MODULES_DIR}"/../Infra_Installer/disaster_recovery/backup-restore-ceph-objectstore.sh -o yaml --dry-run=client | kubectl apply -f -
}

function create_ceph_tmp_backup() {
  local util_image="$1"
  local k8s_node_name="$2"
  local backup_manifest_file="${SCRIPT_DIR}"/k8s_manifests/pre-upgrade-ceph-objectstore-backup-tmp.yaml

  \cp "${SCRIPT_DIR}"/k8s_manifests/pre-upgrade-ceph-objectstore-backup-restore.yaml "${backup_manifest_file}"
  sed -i "s;__IMAGE_PLACEHOLDER__;${util_image};g" "${backup_manifest_file}"
  sed -i "s;__NODE_NAME_PLACEHOLDER__;${k8s_node_name};g" "${backup_manifest_file}"
  sed -i "s;__HOST_PATH_PLACEHOLDER__;${TMP_CEPH_BACKUP_PATH};g" "${backup_manifest_file}"

  kubectl -n "rook-ceph" delete "job/pre-upgrade-ceph-backup" --ignore-not-found
  kubectl apply -f "${backup_manifest_file}"
  kubectl -n "rook-ceph" wait --timeout=30m --for=condition=complete "job/pre-upgrade-ceph-backup" || error "Timeout waiting for ceph tmp backup to complete before downgrade"
}

function cleanup_ceph_resources(){
  local namespace="rook-ceph"
  local k8s_resources_to_clean=(job pod serviceaccount service secret configmap persistentvolumeclaim)
  for resource in "${k8s_resources_to_clean[@]}"; do
    kubectl -n "${namespace}"  get "${resource}" --no-headers -o name  | grep "${resource}" | xargs -r kubectl -n "${namespace}"  patch --type=json -p '[{"op":"remove" , "path": "/metadata/finalizers"}]' || true
    kubectl -n "${namespace}"  get "${resource}" --no-headers -o name  | grep "${resource}" | xargs -r kubectl -n "${namespace}" delete --force --ignore-not-found
  done
}

function create_required_rook_crds() {
  kubectl create -f "${SCRIPT_DIR}"/k8s_manifests/crd_cephbucketnotifications.ceph.rook.io.yaml || true
  kubectl create -f "${SCRIPT_DIR}"/k8s_manifests/crd_cephbuckettopics.ceph.rook.io.yaml || true
  kubectl create -f "${SCRIPT_DIR}"/k8s_manifests/crd_cephfilesystemsubvolumegroups.ceph.rook.io.yaml || true
}

function patch_argocd_rook_apps() {
  local values
  local updated_value

  values=$(kubectl -n argocd get application  rook-ceph-object-store  -o json | jq  '.spec.source.helm.values')
  updated_value=$(echo "$values" | sed  's/16.2.6/15.2.10/g' | sed 's/1.7.9/1.5.9/g' | sed -e "s/^\"//g" -e "s/\"$//g")
  kubectl -n argocd patch  application  rook-ceph-object-store  --type='json' -p='[{"op" : "replace" ,"path" : "/spec/source/helm/values", "value": "'"${updated_value}"'"}]'

  values=$(kubectl -n argocd get application  rook-ceph-operator   -o json | jq  '.spec.source.helm.values')
  updated_value=$(echo "$values"  | sed 's;rook/ceph;sf-rook-ceph-fork\\n  tag: v1.8.9-4;g' | sed 's/1.7.9/1.5.9/g' | sed -e "s/^\"//g" -e "s/\"$//g")
  kubectl -n argocd patch  application  rook-ceph-operator  --type='json' -p='[{"op" : "replace" ,"path" : "/spec/source/helm/values", "value": "'"${updated_value}"'"}]'

  argocd app sync rook-ceph-object-store || true
  argocd app sync rook-ceph-operator || true
}

function restore_ceph_from_tmp_backup() {
  local util_image="$1"
  local k8s_node_name="$2"
  local restore_manifest_file="${SCRIPT_DIR}"/k8s_manifests/pre-upgrade-ceph-objectstore-restore-tmp.yaml

  \cp "${SCRIPT_DIR}"/k8s_manifests/pre-upgrade-ceph-objectstore-backup-restore.yaml "${restore_manifest_file}"
  sed -i "s;__IMAGE_PLACEHOLDER__;${util_image};g" "${restore_manifest_file}"
  sed -i "s;__NODE_NAME_PLACEHOLDER__;${k8s_node_name};g" "${restore_manifest_file}"
  sed -i "s;__HOST_PATH_PLACEHOLDER__;${TMP_CEPH_BACKUP_PATH};g" "${restore_manifest_file}"
  sed -i 's/pre-upgrade-ceph-backup/pre-upgrade-ceph-restore/g'  "${restore_manifest_file}"
  sed -i 's;value: backup;value: restore;g'  "${restore_manifest_file}"
  kubectl -n "rook-ceph" delete "job/pre-upgrade-ceph-restore" --ignore-not-found
  kubectl apply -f "${restore_manifest_file}"
  kubectl -n "rook-ceph" wait --timeout=60m --for=condition=complete "job/pre-upgrade-ceph-restore" || error "Timeout waiting for completion of ceph restore from tmp backup"
}

function backup_ceph() {
  local node_name="$1"
  local sf_util_image="$2"
  local ceph_object_size

  ceph_object_size=$(get_ceph_object_size_bytes)
  validate_ceph_tmp_backup_path "${ceph_object_size}"
  prepare_ceph_tmp_backup_path
  prepare_ceph_tmp_backup_script_configmap
  create_ceph_tmp_backup "${sf_util_image}" "${node_name}"
}

function destroy_ceph() {
  local dsPod
  kubectl -n rook-ceph delete deploy --all
  cleanup_ceph_resources
  for dsPod in $(kubectl -n kube-system get pod -l app=longhorn-loop-device-cleaner -o name); do
    kubectl -n kube-system exec "${dsPod}"  -- rm -rf /var/lib/rook || true
  done
}

function create_ceph_cluster(){
  create_required_rook_crds
  # upgade argocd apps to use downgraded ceph
  patch_argocd_rook_apps
}

function restore_ceph() {
  local node_name="$1"
  local sf_util_image="$2"

  prepare_ceph_tmp_backup_script_configmap
  [[ -n "${TMP_CEPH_BACKUP_PATH}" ]] || error "Ceph restore failed(BackupPathNotProvided). Please provide backup directory (via environment variable 'TMP_CEPH_BACKUP_PATH') for ceph data restore"
  restore_ceph_from_tmp_backup "${sf_util_image}" "${node_name}"
}

function save_ceph_osd_pvc_size() {
  if ! [[ -f "${TMP_CEPH_BACKUP_PATH}/ceph_osd_pvc_size.txt" && -s "${TMP_CEPH_BACKUP_PATH}/ceph_osd_pvc_size.txt" ]]; then
    kubectl -n rook-ceph get pvc -l ceph.rook.io/DeviceSet=set1 -o json | jq -r '.items[].status.capacity.storage' | xargs numfmt --from=auto | sort -r -n | head -n1 > "${TMP_CEPH_BACKUP_PATH}/ceph_osd_pvc_size.txt"
  fi
}

function resize_ceph_osd_pvc() {
  local pvc_size
  [[ -f "${TMP_CEPH_BACKUP_PATH}/ceph_osd_pvc_size.txt" && -s "${TMP_CEPH_BACKUP_PATH}/ceph_osd_pvc_size.txt" ]] || return
  pvc_size=$(numfmt --to=iec-i < "${TMP_CEPH_BACKUP_PATH}/ceph_osd_pvc_size.txt")
  info "Resizing ceph OSD PVC to '${pvc_size}'"
  "${SCRIPT_DIR}/../configureUiPathAS.sh" objectstore resize-pvc -s "${pvc_size}" -b data --force
}

function downgrade_ceph() {
  local node_name
  local sf_util_image
  local is_ceph_downgrade_required=false

  is_ceph_version_stable || is_ceph_downgrade_required=true
  # Set variable when downgrade ot restore is required
  if [[ $is_ceph_downgrade_required == "true" ]] || kubectl -n kube-system get configmap restore-lock >> /dev/null 2>&1; then
    node_name=$(get_local_k8s_node_name)
    [[ -n "${node_name}" ]] || error "Unable to extract kubernetes node name"
    sf_util_image=$(kubectl -n uipath-infra get cronjob istio-configure-script-cronjob  -o jsonpath='{..image}')
    [[ -n "${sf_util_image}" ]] || error "Unable to get sf util image"
    [[ ${INSTALL_TYPE} == 'offline' ]] && sf_util_image="localhost:${DOCKER_REGISTRY_NODEPORT}/${sf_util_image}"
  fi
  # Backup only when ceph is on non stable version
  if [[ $is_ceph_downgrade_required == "true" ]]; then
    save_ceph_osd_pvc_size
    backup_ceph "${node_name}" "${sf_util_image/0.6/0.8}"
    # Disable self heal
    kubectl -n argocd patch application fabric-installer --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":false}]'
    kubectl -n argocd patch application rook-ceph-operator --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":false}]'
    kubectl -n argocd patch application rook-ceph-object-store --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":false}]'
    # Create restore lock to resume the process
    kubectl -n kube-system create configmap restore-lock --dry-run=client -o yaml | kubectl apply -f -
    # Remove ceph cluster
    destroy_ceph
    # Create ceph cluster
    create_ceph_cluster
  fi
  # Wait for ceph cluster to get ready and perform restore when ceph on non stable version or restore lock present
  if [[ $is_ceph_downgrade_required == "true" ]] || kubectl -n kube-system get configmap restore-lock >> /dev/null 2>&1; then
    wait_till_cephobjectstore_ready "rook-ceph" "rook-ceph" "app=rook-ceph-rgw"
    resize_ceph_osd_pvc
    # Run ceph configure job to create ceph admin user
    kubectl -n uipath-infra get job "rook-ceph-configure-script-job" -o json | jq 'del(. | .spec.selector, .spec.template.metadata.labels)' | kubectl replace --force -f -
    kubectl -n uipath-infra wait --timeout=10m --for=condition=complete "job/rook-ceph-configure-script-job"
    kubectl -n "rook-ceph" get secret ceph-object-store-secret --show-labels  | grep  -q 'config-discovery=yes' || error "Ceph Admin credentials not found"
    # Create ceph users
    kubectl -n uipath-infra get job "credential-manager-job" -o json | jq 'del(. | .spec.selector, .spec.template.metadata.labels)' | kubectl replace --force -f -
    kubectl -n uipath-infra wait --timeout=10m --for=condition=complete "job/credential-manager-job"
    restore_ceph "${node_name}" "${sf_util_image/0.6/0.8}"
    kubectl -n kube-system delete configmap restore-lock --ignore-not-found
  fi
}

function check_requirements() {
    local required_pkgs=(kubectl helm)
    check_required_packages_and_commands "${required_pkgs[@]}"
    # validate longhorn disk
    check_longhorn_disks
    # Downgrade ceph
    downgrade_ceph
}

function parse_input_file() {
    VERSION="$(get_optional_json_value ".version")"
    ARGOCD_PASSWORD=""
    FQDN=""
    info "Parsing input file ${INFILE}"
    ARGOCD_PASSWORD="$(get_optional_json_value ".fabric.argocd_admin_password")"
    if [[ -z "${ARGOCD_PASSWORD}" ]]; then
        info "key=\"argocd_admin_password\" not found in ${INFILE}"
        info "Generating new admin password as one was not provided. This will only be used if this is a fresh ArgoCD installation."
        ARGOCD_PASSWORD="$(
            tr </dev/urandom -dc _A-Z-a-z-0-9 | head -c 32
            echo
        )"
    fi
    FQDN="$(get_mandatory_json_value ".fqdn")"
    ARGOCD_FQDN="$(get_mandatory_json_value ".subdomain.argo_cd").${FQDN}"
    ARGOCD_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
    DOCKER_REGISTRY_FQDN="$(get_mandatory_json_value ".subdomain.docker_registry").${FQDN}"
    RANCHER_FQDN="$(get_mandatory_json_value ".subdomain.rancher_server").${FQDN}"
    CEPH_FQDN="$(get_mandatory_json_value ".subdomain.rook_ceph").${FQDN}"
    CEPH_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
    SELF_HEAL_OPERATOR_VERSION="$(get_chart_version_from_manifest "self-heal-operator")"
    UIPATH_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
    DOCKER_REGISTRY_EXTERNAL_PORT="$(get_mandatory_json_value ".cluster_port")"
    FABRIC_CHART="$(get_mandatory_json_value ".fabric.chart")"
    PROFILE="$(get_mandatory_json_value ".profile")"
    HELM_REPO_URL="$(get_mandatory_json_value ".registries.helm.url")"
    HELM_REPO_USERNAME="$(get_optional_json_value ".registries.helm.username")"
    HELM_REPO_PASSWORD="$(get_optional_json_value ".registries.helm.password")"
    HELM_CHART_VERSION="$(get_mandatory_json_value ".version")"
    DOCKER_REGISTRY_URL="$(get_mandatory_json_value ".registries.docker.url")"
    DOCKER_REGISTRY_NODEPORT="$(get_mandatory_json_value ".infra.docker_registry.node_port")"
    DOCKER_REGISTRY_PULL_SECRET="$(get_optional_json_value ".registries.docker.pull_secret_value")"
    STORAGE_CLASS="$(get_mandatory_json_value ".storage_class")"
    STORAGE_CLASS_SINGLE_REPLICA="$(get_mandatory_json_value ".storage_class_single_replica")"
    PROXY_ENABLED="$(get_optional_json_value ".proxy.enabled")"
    HTTP_PROXY="$(get_optional_json_value ".proxy.http_proxy")"
    HTTPS_PROXY="$(get_optional_json_value ".proxy.https_proxy")"
    NO_PROXY="$(get_optional_json_value ".proxy.no_proxy")"
    REDIS_LICENSE="$(get_optional_json_value ".fabric.redis.license")"
    REDIS_HA="$(get_optional_json_value ".fabric.redis.ha")"
    TELEMETRY_OPT_OUT="$(get_optional_json_value ".telemetry_optout")"
    RKE_TOKEN="$(get_mandatory_json_value ".rke_token")"
    PROMETHEUS_MEMORY_LIMIT="$(get_optional_json_value ".monitoring.prometheus_memory_limit")"
    PROMETHEUS_STORAGE_SIZE="$(get_optional_json_value ".monitoring.prometheus_storage_size")"
    PROMETHEUS_RETENTION="$(get_optional_json_value ".monitoring.prometheus_retention")"
    ZONE_RESILIENCE="$(get_json_value_or_return_default_value ".zone_resilience" "false")"
    SELF_HEAL_OPERATOR_OUTPUT_NODE_EVENT="$(get_json_value_or_return_default_value ".infra.self_heal_operator.output_node_event" "false")"
    

    export PROXY_ENABLED
    export DOCKER_REGISTRY_URL
    export PROFILE
    export ZONE_RESILIENCE

    # Disable telemetry in case of offline
    if [[ "${INSTALL_TYPE}" == "offline" ]];
    then
      TELEMETRY_OPT_OUT="true"
    fi
}

function set_output() {
  cat > "${OUTFILE}" <<EOF
{ 
  "fqdn": "${FQDN}",
  "argocd_basepath": "/",
  "argocd_namespace": "argocd",
  "argocd_service_name": "argocd-server",
  "argocd_secret_name": "${ARGOCD_PASSWORD_SECRET_NAME}",
  "argocd_password": "${ARGOCD_PASSWORD}",
  "docker_registry": "${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}",
  "ceph_fqdn": "${CEPH_FQDN}",
  "argocd_fqdn": "${ARGOCD_FQDN}",
  "docker_fqdn": "${DOCKER_REGISTRY_FQDN}",
  "rancher_fqdn": "${RANCHER_FQDN}",
  "argocd_external_port": ${ARGOCD_EXTERNAL_PORT},
  "ceph_external_port": ${CEPH_EXTERNAL_PORT},
  "docker_external_port": ${DOCKER_REGISTRY_EXTERNAL_PORT},
  "rancher_external_port": ${UIPATH_EXTERNAL_PORT},
  "uipath_port": ${UIPATH_EXTERNAL_PORT},
  "sf_image_registry": "${DOCKER_REGISTRY_URL}"
}
EOF
}

########
# Patch argoCD deployments with proxy env variables
# Globals:
# ARGOCD_NAMESPACE, HTTP_PROXY, HTTPS_PROXY, NO_PROXY
########
function patch_proxy_configuration() {
  kubectl -n "${ARGOCD_NAMESPACE}" set env deployment/argocd-repo-server \
      HTTP_PROXY="${HTTP_PROXY}" \
      HTTPS_PROXY="${HTTPS_PROXY}" \
      NO_PROXY="${NO_PROXY}",.svc,.local,.cluster

  kubectl -n "${ARGOCD_NAMESPACE}" set env deployment/argocd-server \
      HTTP_PROXY="${HTTP_PROXY}" \
      HTTPS_PROXY="${HTTPS_PROXY}" \
      NO_PROXY="${NO_PROXY}",127.0.0.0/8,localhost,kubernetes,.svc,.local,.cluster,argocd-repo-server
}

########
# Get/Reset the ArgoCD password
# Globals:
#   ARGOCD_PASSWORD_SECRET_NAME, ARGOCD_NAMESPACE, BINARY_DIR, ARGOCD_BIN_FILENAME,
#   ARGOCD_ENDPOINT
########
function get_or_reset_argocd_password() {
    local argocd_password=${1}
    local secret
    local argocd_default_password
    
    info "Checking if ArgoCD password was reset, looking for secrets/${ARGOCD_PASSWORD_SECRET_NAME}."
    secret=$(kubectl get secret/"${ARGOCD_PASSWORD_SECRET_NAME}" -n "${ARGOCD_NAMESPACE}" --ignore-not-found)
    if [ -z "${secret}" ]
    then
        local try=0
        local maxtry=10
        local status="notready"
        while [[ ${status} == "notready" ]]  && (( try != maxtry )) ; do
                argocd_default_password=$(kubectl -n "${ARGOCD_NAMESPACE}" get secret argocd-initial-admin-secret -o jsonpath="{.data.password}" | base64 -d)
                try=$((try+1))
                argocd login \
                    "${ARGOCD_ENDPOINT}" \
                    --grpc-web --grpc-web-root-path "/" \
                    --username admin --password "${argocd_default_password}" \
                    --insecure \
                    && status="ready"
                info "Secret not found, trying to log in with initial password...${try}/${maxtry}" &&  sleep 20;
        done

        [[ ${status} == "ready" ]] || error_with_code "12001"


        info "Login successful. Resetting password..."

        argocd account update-password \
            --current-password "${argocd_default_password}" \
            --new-password "${argocd_password}" \
            || error_with_code "12002"

        info "Creating secret/${ARGOCD_PASSWORD_SECRET_NAME}"
        info "$(kubectl create secret -n "${ARGOCD_NAMESPACE}" generic "${ARGOCD_PASSWORD_SECRET_NAME}" --from-literal=username=admin --from-literal=password="${argocd_password}")"
        info "To get your new password, run \"kubectl get secrets/${ARGOCD_PASSWORD_SECRET_NAME} -n ${ARGOCD_NAMESPACE} --template {{.data.password}} | base64 -d\""
                kubectl -n "${ARGOCD_NAMESPACE}" delete secret argocd-initial-admin-secret
    else
        info "Secret found, trying to log in..."
        argocd_password=$(kubectl get secrets/"${ARGOCD_PASSWORD_SECRET_NAME}" -n "${ARGOCD_NAMESPACE}" --template \{\{.data.password\}\} | base64 -d)
        argocd login \
            "${ARGOCD_ENDPOINT}" \
            --grpc-web --grpc-web-root-path "/" \
            --username admin --password "${argocd_password}" \
            --insecure || error_with_code "12001"

        info "Login successful!"
        info "To get your password, run kubectl get secrets/${ARGOCD_PASSWORD_SECRET_NAME} -n ${ARGOCD_NAMESPACE} --template \{\{.data.password\}\} | base64 -d"
    fi

    export ARGOCD_PASSWORD=${argocd_password}

    local try=0
    local maxtry=0
    local status="notready"
    while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
            try=$((try+1))
            argocd app list >/dev/null 2>&1  && status="ready"
            info "waiting for argocd to be ready...${try}/${maxtry}" &&  sleep 10;
    done
    argocd app list || error_with_code "12003"
}

function create_argocd_helm() {

  if [ -z "${HELM_REPO_PASSWORD}" ]; then
    cp "${VENDOR_DIR}/argocd/config/argocd-config.public.yaml" "argocd-config.yaml"
  else
    cp "${VENDOR_DIR}/argocd/config/argocd-config.private.yaml" "argocd-config.yaml"
  fi
  sed "s|##HELM_REPOURL##|$(echo -n "${HELM_REPO_URL}"| base64 -w 0)|" -i "argocd-config.yaml"
  sed "s|##HELM_REPOUSERNAME##|$(echo -n "${HELM_REPO_USERNAME}"| base64 -w 0)|" -i "argocd-config.yaml"
  sed "s|##HELM_REPOPASSWORD##|$(echo -n "${HELM_REPO_PASSWORD}"| base64 -w 0)|" -i "argocd-config.yaml"
  kubectl apply -n "${ARGOCD_NAMESPACE}" -f "argocd-config.yaml"

}

function validate_service_installation_type() {
    local service_install_type trigger_service_install_type
    trigger_service_install_type="all"

    service_install_type=$(get_cluster_config_default "SERVICE_INSTALL_TYPE" "all")

    if [[ ${INSTALLSTANDALONEAICENTER} == "1" ]]; then
        trigger_service_install_type='aicenter'
    fi

    if [[ "${service_install_type}" != "${trigger_service_install_type}" ]]; then
      error_with_code "12011"
    fi
}

function send_input_telemetry()  {
  local telemetry_content

  if [[ "${TELEMETRY_OPT_OUT:-false}" == "true" ]]; then
    info "Telemetry is off"
    return;
  fi

  local now
  now=$(date -u +%Y-%m-%dT%H:%M:%S.%3NZ)

  local helmRepoUrl
  if [[ "${INSTALL_TYPE}" == "online" ]]; then
    helmRepoUrl=${HELM_REPO_URL}
  else
    helmRepoUrl="(redacted)"
  fi

  ensure_machine_hash

  function data() {
    cat <<EOF
{
    "name": "AppEvents",
    "time": "$now",
    "iKey": "$TELEMETRY_INSTRUMENTATION_KEY",
    "tags": {
        "ai.cloud.role": "Installer",
        "ai.internal.sdkVersion": "REST"
    },
    "data": {
        "baseType": "EventData",
        "baseData": {
            "ver": 2,
            "name": "Fabric.Input",
            "properties": {
                "ApplicationName": "UiPath.ServiceFabric.Installer",
                "ProductVersion": "${VERSION}",
                "InstallId": "${INSTALL_ID:-none}",
                "InstallType": "${INSTALL_TYPE}",
                "ClusterToken" : "${RKE_TOKEN}",
                "MachineHash" : "${MACHINE_HASH}",
                "FabricChart": "${FABRIC_CHART}",
                "Profile" : "${PROFILE}",
                "HelmRepoUrl": "${helmRepoUrl}",
                "StorageClass": "${STORAGE_CLASS}",
                "StorageClassSingleReplica" : "${STORAGE_CLASS_SINGLE_REPLICA}",
                "ProxyEnabled": "${PROXY_ENABLED}",
                "NoProxy": "${NO_PROXY}",
                "PrometheusMemoryLimit": "${PROMETHEUS_MEMORY_LIMIT}",
                "PrometheusStorageSize": "${PROMETHEUS_STORAGE_SIZE}",
                "PrometheusRetention": "${PROMETHEUS_RETENTION}"
            }
        }
    }
}
EOF
  }

  telemetry_content="$(data)"
  send_telemetry "${telemetry_content}"
}

function check_redis_db_job (){
    local try=0
    local maxtry=3
    local timeout="600s"
    local status="notready"
    local job_label="app.kubernetes.io/component=redis-enterprise-cluster"

    while [[ ${status} == "notready" ]] && [[ ${status} != "failed" ]]  && (( try != maxtry )); do
        try=$((try+1))
        local jobs=""
        info "Looking for the Redis db job...${try}/${maxtry}"
        jobs=$(kubectl -n redis-system get job -l "${job_label}" \
          -o jsonpath='{.items[*].metadata.name}' \
          --ignore-not-found)
        if [[ ${jobs} != "" ]]; then
            info "Checking: ${jobs}...${try}/${maxtry}"  
            kubectl wait -n redis-system --for=condition=complete \
            --timeout="${timeout}" job -l "${job_label}"
            local result=$?
            [[ ${result} == 0 ]] && status="ready"
            [[ ${result} != 0 ]] && status="failed"
        elif (( try ==  maxtry )); then
            status="notfound"
        fi
        sleep 10
    done
    [[ ${status} == "failed" ]] && error_with_code "12006"
    [[ ${status} == "notfound" ]] && info "No Redis db job found"
    { [[ ${status} == "ready" ]] && info "Redis db job succeeded"; } || true
}

function get_additional_parameters() {
    additional_parameters=()

    [ -z "${DOCKER_REGISTRY_PULL_SECRET}" ] || additional_parameters+=(-p global.dockerRegistry.imagePullSecretValue="${DOCKER_REGISTRY_PULL_SECRET}" )

    [ -z "${REDIS_LICENSE}" ] || additional_parameters+=(-p global.redis.license="${REDIS_LICENSE}" )

    # Set Redis ha only if the cluster ha is enabled
    if [[ "${PROFILE}" == "ha" ]] && [[ -n "${REDIS_HA}" ]]; then
      additional_parameters+=(-p global.redis.ha="${REDIS_HA}" )
    fi


    if [[ ${INSTALLSTANDALONEAICENTER} == "1" ]]; then
        additional_parameters+=(-p applications.applications.redis-operator.enabled="false" )
        additional_parameters+=(-p applications.applications.redis-cluster.enabled="false" )
        additional_parameters+=(-p applications.applications.csi-driver-smb.enabled="false" )
    fi
}

function prepare_monitoring_config() {
    # defaults
    if [[ -z "${PROMETHEUS_MEMORY_LIMIT}" ]]; then
      [ "${PROFILE}" == "ha" ] && PROMETHEUS_MEMORY_LIMIT=6000
      [ "${PROFILE}" != "ha" ] && PROMETHEUS_MEMORY_LIMIT=3200
    fi
    [ -z "${PROMETHEUS_STORAGE_SIZE}" ] && PROMETHEUS_STORAGE_SIZE=45
    [ -z "${PROMETHEUS_RETENTION}" ] && PROMETHEUS_RETENTION=7

    # minimums
    if [[ "${PROFILE}" == "ha" ]]; then
      [ "${PROMETHEUS_MEMORY_LIMIT}" -lt 5000 ] && PROMETHEUS_MEMORY_LIMIT=5000
    else
      [ "${PROMETHEUS_MEMORY_LIMIT}" -lt 2500 ] && PROMETHEUS_MEMORY_LIMIT=2500
    fi
    [ "${PROMETHEUS_STORAGE_SIZE}" -lt 45 ] && PROMETHEUS_STORAGE_SIZE=45
    [ "${PROMETHEUS_RETENTION}" -lt 1 ] && PROMETHEUS_RETENTION=1

    PROMETHEUS_RETENTION_STORAGE_LIMIT="$((PROMETHEUS_STORAGE_SIZE-10))"

    # unit suffixes
    PROMETHEUS_MEMORY_LIMIT="${PROMETHEUS_MEMORY_LIMIT}Mi"
    PROMETHEUS_STORAGE_SIZE="${PROMETHEUS_STORAGE_SIZE}Gi"
    PROMETHEUS_RETENTION="${PROMETHEUS_RETENTION}d"
    PROMETHEUS_RETENTION_STORAGE_LIMIT="${PROMETHEUS_RETENTION_STORAGE_LIMIT}GB"
}

# Checks the status of fabric-readiness job, params as follows
# 1 - namespace
# 2 - Job name
function check_fabric_readiness_status() {
  local namespace=$1
  local jobname=$2

  local try=0
  local maxtry=10
  local found="false"
  while [[ ${found} == "false" ]]  && (( try != maxtry )); do
    try=$((try+1))
    local jobs=""
    info "Looking for fabric readiness checks...${try}/${maxtry}"
    jobs=$(kubectl -n "$namespace" get job "$jobname" --ignore-not-found)
    if [[ ${jobs} != "" ]]; then
      info "Waiting for fabric readiness job to complete"
      kubectl -n "$namespace" wait --for=condition=Complete job/"$jobname" --timeout=180s 2>/dev/null
      found="true"
    fi
    sleep 30
  done

  if [[ "${found}" == "false" || "$(kubectl -n "$namespace" get job "$jobname" -o jsonpath='{.status.conditions[?(@.type=="Complete")].status}')" != "True" ]];
  then
    error_with_code "12012" "$namespace" "$namespace" "$jobname"
  else
    info "Fabric readiness checks completed successfully"
  fi
}




function download_self_heal_chart() {
  info "Using local helm folder ${TMP_HELM_CHARTS_DIR}"
  if [[ -n "${HELM_REPO_USERNAME}" ]] && [[ -n "${HELM_REPO_PASSWORD}" ]]; then
    info "Using private Helm registry located at ${HELM_REPO_URL}"
    echo "${HELM_REPO_PASSWORD}" |
      helm registry login "${HELM_REPO_URL}" \
        --username "${HELM_REPO_USERNAME}" \
        --password-stdin &>> "${dev_null}"
  fi

    declare -a charts=(
    "helm/self-heal-operator ${SELF_HEAL_OPERATOR_VERSION}"
  )

  for crt_chart in "${charts[@]}"; do
    IFS=' ' read -ra crt_chart_arr <<<"${crt_chart}"

    local try=0
    local maxtry=10
    local status="notready"
    while [[ ${status} == "notready" ]] && ((try != maxtry)); do
      try=$((try + 1))
      info "Pulling chart ${HELM_REPO_URL}/${crt_chart_arr[0]}:${crt_chart_arr[1]} ...try ${try}/${maxtry}"
      helm chart pull "${HELM_REPO_URL}"/"${crt_chart_arr[0]}":"${crt_chart_arr[1]}" &>> "${dev_null}" && status="ready"

      [[ ${status} == "ready" ]] || sleep 2
    done
    [[ ${status} == "ready" ]] || error_with_code "10022" "${HELM_REPO_URL}/${crt_chart_arr[0]}:${crt_chart_arr[1]}"
    helm chart export "${HELM_REPO_URL}"/"${crt_chart_arr[0]}":"${crt_chart_arr[1]}" --destination "${TMP_HELM_CHARTS_DIR}" &>> "${dev_null}"
  done


}

function setup_helm_directory() {
  if [[ ${INSTALL_TYPE} == "offline" ]]; then
    info "Using private Helm registry located at ${HELM_REPO_URL}"
    echo "${HELM_REPO_PASSWORD}" |
      helm registry login "${HELM_REPO_URL}" \
        --username "${HELM_REPO_USERNAME}" \
        --password-stdin &>> "${dev_null}"

    local try=0
    local maxtry=10
    local status="notready"
    while [[ ${status} == "notready" ]] && ((try != maxtry)); do
      try=$((try + 1))
      info "Exporting chart ${HELM_REPO_URL}/${SELF_HEAL_CHART} ...try ${try}/${maxtry}"
      helm chart export "${HELM_REPO_URL}"/"${SELF_HEAL_CHART}":"${SELF_HEAL_OPERATOR_VERSION}" --destination "${TMP_HELM_CHARTS_DIR}" &>> "${dev_null}" && status="ready"

      [[ ${status} == "ready" ]] || sleep 2
    done
    [[ ${status} == "ready" ]] || error_with_code "10022" "${HELM_REPO_URL}/${SELF_HEAL_CHART}"  

  else
    TMP_HELM_CHARTS_DIR=$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t sf-selfheal-helm-chart.XXXXXXXXXX)
    download_self_heal_chart
  fi
  info "Using infra helm charts from ${TMP_HELM_CHARTS_DIR}"
  export TMP_HELM_CHARTS_DIR
}


function install_self_heal_operator() {
  local infra_namespace="uipath-infra"
  info "Installing self-heal-operator"
  local INSTALL_SELF_HEAL_OPERATOR_ARGS=()

  [[ "${INSTALL_TYPE}" == "online" ]] && INSTALL_SELF_HEAL_OPERATOR_ARGS+=(--set image.registry="${DOCKER_REGISTRY_URL}")
  [[ -n "${DOCKER_REGISTRY_PULL_SECRET}" ]] && INSTALL_SELF_HEAL_OPERATOR_ARGS+=(--set "imagePullSecrets[0].name=uipathpullsecret")

  INSTALL_SELF_HEAL_OPERATOR_ARGS+=(--set zoneResilient="${ZONE_RESILIENCE}")
  INSTALL_SELF_HEAL_OPERATOR_ARGS+=(--set outputNodeEvent="${SELF_HEAL_OPERATOR_OUTPUT_NODE_EVENT}")

  helm upgrade  --install --atomic self-heal-operator "${TMP_HELM_CHARTS_DIR}"/self-heal-operator/ \
    --wait --timeout 10m \
    --namespace $infra_namespace \
    --create-namespace \
    "${INSTALL_SELF_HEAL_OPERATOR_ARGS[@]}" &>/dev/null || {
      error_without_exit "${LANG_SELF_HEAL_FAILED_PREFIX}";
      error_with_code "11031";
    }
  info "Finished Installing self-heal-operator"
}


function main() {
    local try
    local maxtry
    local status
    local old_trap

    cd "${SCRIPT_DIR}"

    info "${LANG_FABRIC_STARTED_PREFIX} ${LANG_FABRIC_ALM_SETUP_MESSAGE}"
    # Use trap to append extra error messages to functions imported in this script
    old_trap="$(get_existing_trap "EXIT")"
    #shellcheck disable=SC2064
    trap "error_without_exit \"${LANG_FABRIC_FAILED_PREFIX} ${LANG_FABRIC_ALM_SETUP_MESSAGE}\"; ${old_trap}" EXIT

    UPDATE_FQDN=$(get_cluster_config "update_fqdn" )
    if [[ "${UPDATE_FQDN}" == "true" ]];
    then
      fqdn_update_state=$(get_cluster_config "update_fqdn_state")
      { [[ ${fqdn_update_state} == "" ]] || [[ ${fqdn_update_state} == "INFRA_STARTED" ]]; } \
        && error_with_code "13006"
      update_cluster_config "update_fqdn_state" "FABRIC_STARTED"
    fi

    parse_args "$@"
    parse_input_file
    prepare_monitoring_config

    send_checkpoint_telemetry "Fabric: Input Parsed"
    send_input_telemetry

    if [[ ${INSTALL_TYPE} == 'offline' ]];then
        HELM_REPO_URL="${DOCKER_REGISTRY_FQDN}:${DOCKER_REGISTRY_EXTERNAL_PORT}"
        DOCKER_REGISTRY_URL="localhost:${DOCKER_REGISTRY_NODEPORT}"
    fi

    get_fabric_install_paths "${SCRIPT_DIR}"
    
    load_os_utils "${MODULES_DIR}"
    check_requirements

    send_checkpoint_telemetry "Fabric: Requirements Checked"

    check_available_masters

    try=0
    maxtry=30
    status="notready"
    while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
            try=$((try+1))
            kubectl cluster-info >/dev/null 2>&1  && status="ready"
            info "waiting for cluster to be ready...${try}/${maxtry}" &&  sleep 10;
    done
    kubectl cluster-info || error_with_code "10004"

    send_checkpoint_telemetry "Fabric: Cluster Ready"

    validate_service_installation_type


    if [[ ${INSTALL_TYPE} == "offline" ]]; then
      HELM_REPO_PASSWORD=$(kubectl get secrets/docker-registry-credentials -n "uipath" --template \{\{.data.password\}\} | base64 -d)
      HELM_REPO_USERNAME=$(kubectl get secrets/docker-registry-credentials -n "uipath" --template \{\{.data.username\}\} | base64 -d)
    fi

    # Moving self heal operator installtion to fabric install layer
    setup_helm_directory
    install_self_heal_operator

    get_argocd_install_params

    install_or_update_argocd



    create_argocd_helm

    local ARGOCD_ENDPOINT="${ARGOCD_FQDN}:${ARGOCD_PORT}"
    export ARGOCD_ENDPOINT

    get_or_reset_argocd_password "${ARGOCD_PASSWORD}"

    if [[ ${INSTALL_TYPE} == "offline" ]]; then
        update_argocd_trusted_certs "${DOCKER_REGISTRY_FQDN}"
    fi

    send_checkpoint_telemetry "Fabric: Argo Installed"
    info "${LANG_FABRIC_COMPLETED_PREFIX} ${LANG_FABRIC_ALM_SETUP_MESSAGE}"

    info "${LANG_FABRIC_STARTED_PREFIX} ${LANG_FABRIC_CLUSTER_TOOLS_SETUP_MESSAGE}"
    #shellcheck disable=SC2064
    trap "error_without_exit \"${LANG_FABRIC_FAILED_PREFIX} ${LANG_FABRIC_CLUSTER_TOOLS_SETUP_MESSAGE}\"; ${old_trap}" EXIT
    
    VALUES_YAML="values.yaml"
    if [ "${PROFILE}" != "default" ]; then
      VALUES_YAML="values-${PROFILE}.yaml"
    fi

    get_additional_parameters

    try=0
    maxtry=5
    status="notready"
    local fabric_app_name="fabric-installer"
    while [[ ${status} == "notready" ]] && (( try != maxtry )) ; do
        try=$((try+1))
        argocd app create ${fabric_app_name} \
            --repo "${HELM_REPO_URL}" \
            --helm-chart "helm/${FABRIC_CHART}" \
            --dest-namespace default \
            --dest-server https://kubernetes.default.svc \
            --revision "${HELM_CHART_VERSION}" \
            --helm-version "v3" \
            --sync-policy "automated" \
            --sync-retry-backoff-duration 30s \
            --sync-retry-backoff-factor 2 \
            --sync-retry-backoff-max-duration 5m \
            --sync-retry-limit 10 \
            "${additional_parameters[@]}" \
            -p global.hostname="${FQDN}" \
            -p global.cephHostname="${CEPH_FQDN}" \
            -p global.cephExternalPort="${CEPH_EXTERNAL_PORT}" \
            -p global.dockerRegistry.url="${DOCKER_REGISTRY_URL}" \
            -p global.installType="${INSTALL_TYPE}" \
            -p global.helm.url="${HELM_REPO_URL}" \
            -p global.storageClass="${STORAGE_CLASS}" \
            -p global.storageClassSingleReplica="${STORAGE_CLASS_SINGLE_REPLICA}" \
            -p global.profile="${PROFILE}" \
            -p global.rancherMonitoring.fqdn="${RANCHER_FQDN}" \
            -p global.rancherMonitoring.prometheus.memoryLimit="${PROMETHEUS_MEMORY_LIMIT}" \
            -p global.rancherMonitoring.prometheus.storage.volumeClaimTemplate.spec.resources.requests.storage="${PROMETHEUS_STORAGE_SIZE}" \
            -p global.rancherMonitoring.prometheus.retention="${PROMETHEUS_RETENTION}" \
            -p global.rancherMonitoring.prometheus.retentionSize="${PROMETHEUS_RETENTION_STORAGE_LIMIT}" \
            -p global.zoneResilience="${ZONE_RESILIENCE}" \
            --values "${VALUES_YAML}" \
            --upsert > /dev/null\
            && status="ready"

        send_checkpoint_telemetry "Fabric: Argo App Create try ${try}/${maxtry}"
        info "Creating the fabric ${fabric_app_name} app...try ${try}/${maxtry}" &&  sleep 20;
    done
    [[ ${status} == "ready" ]] || error_with_code "12007" "${fabric_app_name}"
    send_checkpoint_telemetry "Fabric: Syncing the ${fabric_app_name} app"
    info "Syncing the ${fabric_app_name} app"

    sync_with_wait "${fabric_app_name}" argocd
    
    # We deploy the DB through a post sync job and we need to check it's status
    if [[ ${INSTALLSTANDALONEAICENTER} != "1" ]]; then
        check_redis_db_job
    fi

    check_fabric_readiness_status uipath fabric-readiness-tests

    # TODO remove this when the docker registry will be exposed via Istio
    # DOCKER_NODE_PORT=$(kubectl -n docker-registry get service docker-registry -o jsonpath='{.spec.ports[?(@.port==5000)].nodePort}')

    set_output

    [[ "${UPDATE_FQDN}" == "true" ]] && update_cluster_config "update_fqdn_state" "FABRIC_DONE"
    
    send_checkpoint_telemetry "Fabric: Success"
    #shellcheck disable=SC2064
    trap "${old_trap}" EXIT
    info "${LANG_FABRIC_COMPLETED_PREFIX} ${LANG_FABRIC_CLUSTER_TOOLS_SETUP_MESSAGE}"
    info "Fabric Components installed successfully."

}

main "$@"
